try:
    from smolagents.tools import Tool as SmolTool
    from smolagents.tools import tool as smol_tool_decorator

    _HAS_SMOLAGENTS = True
except ImportError:
    _HAS_SMOLAGENTS = False


class SmolAgentsMixin:
    def import_from_smolagents(self, smol_tool):
        if not _HAS_SMOLAGENTS:
            raise RuntimeError(
                "smolagents is not installed or could not be imported. "
                "Install it or check your environment."
            )
        self.name = smol_tool.name[0]
        self.description = smol_tool.description

        def adapted_fn(*args, **kwargs):
            return smol_tool.forward(*args, **kwargs)

        self.fn = adapted_fn

    def export_to_smolagents(self):
        if not _HAS_SMOLAGENTS:
            raise RuntimeError(
                "smolagents is not installed or could not be imported. "
                "Install it or check your environment."
            )

        def smol_forward(*args, **kwargs):
            return self.fn(*args, **kwargs)

        # Get the parameters from the Gofannon tool definition
        parameters = self.definition.get("function", {}).get("parameters", {})
        param_properties = parameters.get("properties", {})
        required_params = parameters.get("required", [])

        # Create inputs_definition from the Gofannon tool definition
        inputs_definition = {}
        for param_name, param_def in param_properties.items():
            inputs_definition[param_name] = {
                "type": param_def.get("type", "string"),
                "description": param_def.get("description", ""),
                #"required": param_name in required_params
            }

            # Get the description from the Gofannon tool definition
        description = self.definition.get("function", {}).get("description", "Exported from Gofannon tool")

        exported_tool = SmolTool()
        exported_tool.name = getattr(self, "name", "exported_base_tool")
        exported_tool.description = description
        exported_tool.inputs = inputs_definition
        exported_tool.output_type = "string"
        exported_tool.forward = smol_forward
        exported_tool.is_initialized = True

        return exported_tool


import os

from..base import BaseTool
from ..config import FunctionRegistry
import logging

logger = logging.getLogger(__name__)

@FunctionRegistry.register
class ListDirectory(BaseTool):
    def __init__(self, name="list_directory"):
        super().__init__()
        self.name = name

    @property
    def definition(self):
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": "List the contents of a directory recursively in a tree-like format",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "directory_path": {
                            "type": "string",
                            "description": "The path of the directory to list"
                        },
                        "max_depth": {
                            "type": "integer",
                            "description": "Maximum depth to recurse into (default: 5)",
                            "default": 5
                        }
                    },
                    "required": ["directory_path"]
                }
            }
        }

    def _build_tree(self, path, prefix="", depth=0, max_depth=5):
        if depth > max_depth:
            return ""

        try:
            entries = os.listdir(path)
        except PermissionError:
            return f"{prefix}[Permission Denied]\n"
        except FileNotFoundError:
            return f"{prefix}[Directory Not Found]\n"

        tree = ""
        entries.sort()
        length = len(entries)

        for i, entry in enumerate(entries):
            full_path = os.path.join(path, entry)
            is_last = i == length - 1

            if os.path.isdir(full_path):
                tree += f"{prefix}{'└── ' if is_last else '├── '}{entry}/\n"
                tree += self._build_tree(
                    full_path,
                    prefix + ("    " if is_last else "│   "),
                    depth + 1,
                    max_depth
                )
            else:
                tree += f"{prefix}{'└── ' if is_last else '├── '}{entry}\n"

        return tree

    def fn(self, directory_path, max_depth=5):
        logger.debug(f"Listing directory: {directory_path}")

        if not os.path.exists(directory_path):
            return f"Error: Directory '{directory_path}' does not exist"

        if not os.path.isdir(directory_path):
            return f"Error: '{directory_path}' is not a directory"

        tree = self._build_tree(directory_path, max_depth=max_depth)
        return f"{directory_path}/\n{tree}"


# gofannon/base/llamastack.py

import inspect
import re
from textwrap import dedent
from typing import Callable, Any

# Type mapping from Gofannon definition to Python types (primarily for potential future use,
# as Llama Stack example doesn't heavily rely on type hints in the generated function)
GOFANNON_TO_PYTHON_TYPE_MAP = {
    "string": str,
    "number": float,
    "integer": int,
    "boolean": bool,
    "object": dict,
    "array": list,
}

# Type mapping from Python types/common hints to Gofannon definition types
PYTHON_TO_GOFANNON_TYPE_MAP = {
    str: "string",
    float: "number",
    int: "integer",
    bool: "boolean",
    dict: "object",
    list: "array",
    Any: "string", # Default fallback
    type(None): "string" # Default fallback for untyped
}


class LlamaStackMixin:
    """
    Mixin for converting Gofannon tools to and from the Llama Stack custom tool format.

    Llama Stack expects custom tools as Python functions with specific docstrings.
    See: https://github.com/meta-llama/llama-stack/blob/main/docs/tools.md#adding-custom-tools
    """

    def _parse_llamastack_docstring(self, docstring: str) -> tuple[str, dict[str, str], list[str]]:
        """
        Parses a Llama Stack style docstring to extract description and parameters.

        Args:
            docstring: The docstring content.

        Returns:
            A tuple containing:
            - description (str): The main tool description.
            - params (dict): A dictionary mapping parameter names to their descriptions.
            - required (list): A list of required parameter names (inferred if not explicitly optional).
              NOTE: Llama Stack's doc example doesn't show optional syntax, so we assume all
              documented params are required unless function signature has defaults.
        """
        if not docstring:
            return "No description provided.", {}, []

        lines = docstring.strip().split('\n')
        description_lines = []
        params = {}
        param_section_started = False

        # Simple regex to find ':param <name>:' or ':param <name> (<type>):'
        param_regex = re.compile(r":param\s+([\w_]+)(?:\s*\([^)]+\))?:\s*(.*)")

        for line in lines:
            line = line.strip()
            if not line: # Skip empty lines between description and params
                if description_lines and not param_section_started:
                    param_section_started = True # Assume blank line separates desc from params
                continue

            match = param_regex.match(line)
            if match:
                param_section_started = True
                param_name = match.group(1)
                param_desc = match.group(2).strip()
                params[param_name] = param_desc
            elif not param_section_started:
                description_lines.append(line)

        description = " ".join(description_lines).strip()
        # Llama Stack example doesn't specify required/optional in docstring.
        # We'll infer from function signature later.
        # For now, return an empty required list based purely on docstring parsing.
        required = list(params.keys()) # Assume all documented params are required initially

        return description, params, required

    def import_from_llamastack(self, llamastack_tool_func: Callable):
        """
        Adapts a Llama Stack custom tool function to a Gofannon tool.

        Args:
            llamastack_tool_func: The Python function representing the Llama Stack tool.
        """
        if not callable(llamastack_tool_func):
            raise ValueError("Input must be a callable function.")

        self.name = llamastack_tool_func.__name__
        docstring = inspect.getdoc(llamastack_tool_func) or ""

        # Parse docstring for description and param descriptions
        description, param_descriptions, _ = self._parse_llamastack_docstring(docstring)
        self.description = description # Set top-level description

        # Use inspect.signature to get parameter names, types, and defaults
        try:
            sig = inspect.signature(llamastack_tool_func)
            parameters_properties = {}
            required_params = []

            for name, param in sig.parameters.items():
                param_type_hint = param.annotation
                gofannon_type = PYTHON_TO_GOFANNON_TYPE_MAP.get(param_type_hint, "string") # Default to string if unknown

                parameters_properties[name] = {
                    "type": gofannon_type,
                    "description": param_descriptions.get(name, f"Parameter '{name}'") # Use parsed desc or default
                }

                # If the parameter has no default value, it's required
                if param.default is inspect.Parameter.empty:
                    required_params.append(name)

            self.definition = {
                "function": {
                    "name": self.name,
                    "description": self.description,
                    "parameters": {
                        "type": "object",
                        "properties": parameters_properties,
                        "required": required_params,
                    }
                }
            }

        except Exception as e:
            # Fallback if signature inspection fails (e.g., built-in function)
            # Create a basic definition based only on name/docstring
            self.logger.warning(f"Could not inspect signature for {self.name}: {e}. Creating basic definition.")
            self.definition = {
                "function": {
                    "name": self.name,
                    "description": self.description,
                    "parameters": { # Assume no parameters if signature fails
                        "type": "object",
                        "properties": {},
                        "required": []
                    }
                }
            }


            # The core function is the Llama Stack function itself
        self.fn = llamastack_tool_func

        self.logger.info(f"Imported Llama Stack tool: {self.name}")


    def export_to_llamastack(self) -> Callable:
        """
        Convert the Gofannon tool to a Llama Stack compatible function.

        Returns:
            A callable function with a Llama Stack-style docstring.
        """
        gofannon_def = self.definition.get("function", {})
        tool_name = gofannon_def.get("name", getattr(self, "name", "gofannon_exported_tool"))
        tool_description = gofannon_def.get("description", getattr(self, "description", "No description provided."))
        parameters = gofannon_def.get("parameters", {})
        param_properties = parameters.get("properties", {})
        required_params = parameters.get("required", [])

        # Construct the docstring
        docstring_lines = [tool_description, ""]
        param_lines = []
        arg_names = []

        for param_name, param_def in param_properties.items():
            param_desc = param_def.get("description", "")
            # Llama Stack example doesn't show type in docstring, just param name and description
            # param_type = param_def.get("type", "string") # Could potentially add :type if needed
            docstring_line = f":param {param_name}: {param_desc}"
            # NOTE: Llama Stack example doesn't explicitly mark required/optional in docstring
            # We could add (required) or (optional) if desired, but sticking to example format.
            # if param_name in required_params:
            #     docstring_line += " (required)"
            param_lines.append(docstring_line)
            arg_names.append(param_name)

        if param_lines:
            docstring_lines.extend(param_lines)
            docstring_lines.append("") # Add blank line after params if any

        # Add a basic return description (Gofannon definition doesn't store this explicitly)
        docstring_lines.append(":return: The result of executing the tool.")

        final_docstring = "\n".join(docstring_lines)

        # Create the wrapper function dynamically
        original_fn = self.fn
        args_string = ", ".join(arg_names)

        # We need to define the function in a scope where original_fn is accessible
        # Using exec can be risky, but is one way to dynamically create a function
        # with a specific signature and docstring. A safer alternative might involve
        # function factories or functools.wraps if the signature complexity allows.

        # Let's try a closure approach which is generally safer:
        def make_wrapper(original_func, doc, name, signature_args):
            # Construct the function signature string dynamically
            sig_str = f"({', '.join(signature_args)})"
            # Use eval to create the function with the correct signature
            # This is still somewhat risky, ensure signature_args are sanitized if needed.
            # Define the wrapper within this factory function's scope
            def wrapper(*args, **kwargs):
                # Map positional args to keywords if necessary, or rely on kwargs
                call_kwargs = {}
                if args:
                    for i, arg_val in enumerate(args):
                        if i < len(signature_args):
                            call_kwargs[signature_args[i]] = arg_val
                        else:
                            # Handle extra positional args if necessary, maybe raise error?
                            pass
                call_kwargs.update(kwargs)
                return original_func(**call_kwargs)

            wrapper.__doc__ = doc
            wrapper.__name__ = name
            # Try to mimic signature for inspection tools (might not be perfect)
            try:
                # Build parameter list for inspect.Signature
                params = [inspect.Parameter(arg, inspect.Parameter.POSITIONAL_OR_KEYWORD) for arg in signature_args]
                wrapper.__signature__ = inspect.Signature(parameters=params)
            except Exception as e:
                self.logger.warning(f"Could not set __signature__ for {name}: {e}")

            return wrapper

            # Generate the arguments list for the signature
        signature_args_list = list(param_properties.keys())
        exported_function = make_wrapper(original_fn, final_docstring, tool_name, signature_args_list)


        # Alternative using exec (use with caution):
        # func_code = f"""
        # def {tool_name}({args_string}):
        #     '''{final_docstring}'''
        #     # Prepare kwargs for the original function
        #     kwargs_for_original = {{}}
        # """
        # for arg_name in arg_names:
        #     func_code += f"    kwargs_for_original['{arg_name}'] = {arg_name}\n"
        #
        # func_code += f"""
        #     return original_fn(**kwargs_for_original)
        # """
        # local_scope = {'original_fn': original_fn}
        # exec(dedent(func_code), local_scope)
        # exported_function = local_scope[tool_name]

        self.logger.info(f"Exported Gofannon tool '{self.name}' to Llama Stack format as function '{tool_name}'")
        return exported_function
  



from..base import BaseTool
import requests
from ..config import FunctionRegistry
import logging

logger = logging.getLogger(__name__)

@FunctionRegistry.register
class Search(BaseTool):
    def __init__(self, name="search"):
        super().__init__()
        self.name = name

    @property
    def definition(self):
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": "Search for articles on arXiv",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "query": {
                            "type": "string",
                            "description": "The search query"
                        },
                        "start": {
                            "type": "integer",
                            "description": "The start index of the search results"
                        },
                        "max_results": {
                            "type": "integer",
                            "description": "The maximum number of search results to return"
                        },
                        "submittedDateFrom": {
                            "type": "string",
                            "description": "The start submission date in the format YYYYMMDD"
                        },
                        "submittedDateTo": {
                            "type": "string",
                            "description": "The end submission date in the format YYYYMMDD"
                        },
                        "ti": {
                            "type": "string",
                            "description": "Search in title"
                        },
                        "au": {
                            "type": "string",
                            "description": "Search in author"
                        },
                        "abs": {
                            "type": "string",
                            "description": "Search in abstract"
                        },
                        "co": {
                            "type": "string",
                            "description": "Search in comment"
                        },
                        "jr": {
                            "type": "string",
                            "description": "Search in journal reference"
                        },
                        "cat": {
                            "type": "string",
                            "description": "Search in subject category"
                        }
                    },
                    "required": ["query"]
                }
            }
        }

    def _format_date(self, date):
        if len(date) == 8:
            return f"{date}0000"
        return date

    def fn(self, query, start=0, max_results=10, submittedDateFrom=None, submittedDateTo=None, ti=None, au=None, abs=None, co=None, jr=None, cat=None):
        logger.debug("Querying ArXiv for '%s'", query)
        base_url = "http://export.arxiv.org/api/query"
        params = {
            "search_query": query,
            "start": start,
            "max_results": max_results
        }

        if submittedDateFrom and submittedDateTo:
            params["search_query"] += f" AND submittedDate:[{self._format_date(submittedDateFrom)}0000 TO {self._format_date(submittedDateTo)}0000]"
        elif submittedDateFrom:
            params["search_query"] += f" AND submittedDate:[{self._format_date(submittedDateFrom)}0000 TO *]"
        elif submittedDateTo:
            params["search_query"] += f" AND submittedDate:[* TO {self._format_date(submittedDateTo)}0000]"

        if ti:
            params["search_query"] += f" AND ti:{ti}"

        if au:
            params["search_query"] += f" AND au:{au}"

        if abs:
            params["search_query"] += f" AND abs:{abs}"

        if co:
            params["search_query"] += f" AND co:{co}"

        if jr:
            params["search_query"] += f" AND jr:{jr}"

        if cat:
            params["search_query"] += f" AND cat:{cat}"

        response = requests.get(base_url, params=params)
        return response.text



from..base import BaseTool
import requests
import json
from ..config import FunctionRegistry
import logging

logger = logging.getLogger(__name__)

@FunctionRegistry.register
class ReadIssue(BaseTool):
    def __init__(self, api_key=None, name="read_issue"):
        super().__init__()
        self.api_key = api_key
        self.name = name
        self.API_SERVICE = 'github'

    @property
    def definition(self):
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": "Read an issue and its comments from a GitHub repository",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "repo_url": {
                            "type": "string",
                            "description": "The URL of the repository, e.g. https://github.com/The-AI-Alliance//gofannon"
                        },
                        "issue_number": {
                            "type": "integer",
                            "description": "The number of the issue to read"
                        }
                    },
                    "required": ["repo_url", "issue_number"]
                }
            }
        }

    def fn(self, repo_url, issue_number):
        logger.debug(f"Reading issue number {issue_number} from repo {repo_url}")
        # Extracting the owner and repo name from the URL
        repo_parts = repo_url.rstrip('/').split('/')
        owner = repo_parts[-2]
        repo = repo_parts[-1]

        issue_url = f"https://api.github.com/repos/{owner}/{repo}/issues/{issue_number}"
        comment_url = f"https://api.github.com/repos/{owner}/{repo}/issues/{issue_number}/comments"

        headers = {
            'Authorization': f'token {self.api_key}'
        }

        issue_response = requests.get(issue_url, headers=headers)
        issue_response.raise_for_status()

        comment_response = requests.get(comment_url, headers=headers)
        comment_response.raise_for_status()

        issue_data = issue_response.json()
        comment_data = comment_response.json()

        result = {
            "issue": issue_data,
            "comments": comment_data
        }

        return json.dumps(result, indent=4)  


from typing import Type, Callable
import subprocess
import shutil
import zipfile
import json
import time

try:
    import boto3
    from botocore.exceptions import ClientError
    from botocore.client import BaseClient

    _HAS_BOTO3 = True
except ImportError:
    _HAS_BOTO3 = False

try:
    from jsonschema import validate
    import jsonschema.exceptions

    _HAS_JSONSCHEMA = True
except ImportError:
    _HAS_JSONSCHEMA = False


class BedrockMixin:

    def export_to_bedrock(self, agent_app_config: dict = None) -> dict:
        """
        Export tool as Bedrock Agent tool configuration
        """
        if not _HAS_BOTO3:
            error = "boto3 not installed. Install with `pip install boto3`"
            self.logger.error(error)
            raise RuntimeError(error)

        if not _HAS_JSONSCHEMA:
            error = "jasonschema not installed. Install with `pip install jsonschema`"
            self.logger.error(error)
            raise RuntimeError(error)

        valid_agent_app_config_schema = {
            "type": "object",
            "properties": {
                "app_id": {"type": "string"},
                "agent_session_timeout": {"type": "integer"},
                "instruction": {"type": "string"},
                "agent_description": {"type": "string"},
                "target_model": {"type": "string"},
                "python_runtime_version": {"type": "string"},
                "temp_build_root": {"type": "string"},
            },
            "required": [
                "app_id",
                "agent_session_timeout",
                "instruction",
                "agent_description",
                "target_model",
                "python_runtime_version",
                "temp_build_root",
            ],
        }
        try:
            validate(agent_app_config, valid_agent_app_config_schema)
        except jsonschema.exceptions.ValidationError as err:
            error = f"JSON validation failure on input parameters: {err}"
            self.logger.error(error)
            raise RuntimeError(error)

        self.bedrock_agent_client = boto3.client("bedrock-agent")
        self.app_id = agent_app_config["app_id"]
        self.agent_name = agent_app_config["agent_name"]
        self.agent_session_timeout = agent_app_config["agent_session_timeout"]
        self.agent_instruction = agent_app_config["instruction"]
        self.agent_description = agent_app_config["agent_description"]
        self.agent_target_model = agent_app_config["target_model"]
        self.python_runtime_version = agent_app_config["python_runtime_version"]
        self.temp_build_root = agent_app_config["temp_build_root"]
        try:
            self.aws_account_id = (
                boto3.client("sts").get_caller_identity().get("Account")
            )
        except ClientError as e:
            error = f"Error getting AWS account number. Client error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)
        except Exception as e:
            error = f"Error getting AWS account number. Unexpected error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        self.openapi_schema_dict = self._generate_openapi_schema()

        self.logger.info(
            f"Starting build of app {self.app_id}...", self.__class__.__name__
        )
        # Creates:
        # the lambda - a .zip archive that contains all the dependencies and the lambda source.
        #              This lambda will have a resource-based policy attached giving it
        #              permission to be invoked by the agent attached.
        # an IAM role - the lambda execution role with managed attached policy
        #               AWSLambdaBasicExecutionRole allowing the lambda to use CloudWatch

        self.logger.info("\tCreating lambda...", self.__class__.__name__)

        self.lambda_arn = self._create_bedrock_lambda()

        # Creates:
        # the Bedrock agent - The bedrock agent itself
        # an IAM role - The agent execution role
        # An IAM policy - a policy to attach to the role allowing the agent to invoke the
        #               - foundational model
        self.logger.info("\tCreating agent...", self.__class__.__name__)
        self.agent_id = self._create_bedrock_agent()

        # Creates:
        # the Bedrock agent action group - Defined with an API schema
        self.logger.info("\tCreating agent action group...", self.__class__.__name__)
        action_group = self._create_agent_action_group()

        # Finally, prepare the agent
        self.logger.info("\tPreparing agent...", self.__class__.__name__)
        try:
            response = self.bedrock_agent_client.prepare_agent(agentId=self.agent_id)
        except ClientError as e:
            error = f"Error preparing agent. Client error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)
        except Exception as e:
            error = f"Error preparing agent. Unexpected error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        output_manifest = {
            "lambdaARN": self.lambda_arn,
            "lambdaRoleName": self.lambda_role_name,
            "agentId": self.agent_id,
            "agentRoleName": self.agent_role_name,
            "agentPolicyARN": self.agent_role_policy_arn,
            "agentActionGroup": self.agent_action_group_id,
        }
        self.logger.info("Done!", self.__class__.__name__)
        return output_manifest

    def _generate_openapi_schema(self) -> dict:
        """Convert Gofannon definition to OpenAPI schema"""
        params = self.definition["function"]["parameters"]
        if not "properties" in params:
            params["properties"] = {}
        openapi_schema = {
            "openapi": "3.0.0",
            "info": {"title": self.name, "version": "1.0.0"},
            "paths": {
                f"/{self.name}": {
                    "get": {
                        "description": self.definition["function"]["description"],
                        "requestBody": {
                            "content": {
                                "application/json": {
                                    "schema": {
                                        "type": "object",
                                        "properties": {
                                            param: {
                                                "type": props["type"],
                                                "description": props["description"],
                                            }
                                            for param, props in params[
                                                "properties"
                                            ].items()
                                        },
                                        "required": params.get("required", []),
                                    }
                                }
                            }
                        },
                        "responses": {
                            "200": {
                                "description": "Successful operation",
                                "content": {
                                    "application/json": {
                                        "schema": {
                                            "type": "object",
                                            "properties": {
                                                "result": {"type": "string"}
                                            },
                                        }
                                    }
                                },
                            }
                        },
                    }
                }
            },
        }

        return openapi_schema

    def _create_bedrock_lambda(self) -> str:
        """Create Lambda function for Bedrock integration"""

        lambda_client = boto3.client("lambda")
        self.lambda_role_arn = self._create_bedrock_lambda_role()

        # Build the lambda .zip deployment package here. Package consits of:
        # 1. All the Python dependencies, including Gofannon itself
        # 2. The lambda source code.
        # bash code as comments at the bottom of this file.
        # See: https://docs.aws.amazon.com/lambda/latest/dg/python-package.html

        # TODO: verify the path here
        command = f"""../../scripts/build_lambda.sh {self.temp_build_root} {self.python_runtime_version.replace("python","")} {self.app_id}"""
        try:
            process = subprocess.Popen(
                command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            stdout, stderr = process.communicate()

            if process.returncode == 0:
                self.built_archive = stdout.decode("utf-8").replace("\n", "")
            else:
                raise RuntimeError(stderr.decode("utf-8"))

        except Exception as e:
            error = f"Error creating .zip archive. {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        # TODO: Skeleton code in the function directory or here?
        lambda_src_string = self._get_lambda_source(
            module_name=self.__class__.__module__, class_name=self.__class__.__name__
        )

        try:
            with open(f"{self.temp_build_root}/lambda_function.py", "w") as text_file:
                text_file.write(lambda_src_string)
        except Exception as e:
            error = f"Error writing lambda Python source file: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        try:
            with zipfile.ZipFile(
                self.built_archive,
                "a",
                compression=zipfile.ZIP_DEFLATED,
            ) as zipf:
                zipf.write(
                    f"{self.temp_build_root}/lambda_function.py", "lambda_function.py"
                )
        except Exception as e:
            error = f"Error inserting lambda Python source file into zip: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        with open(self.built_archive, "rb") as f:
            zipped_code = f.read()

        try:
            response = lambda_client.create_function(
                FunctionName=f"{self.app_id}_{self.name}",
                Runtime=self.python_runtime_version,
                Role=self.lambda_role_arn,
                Handler="lambda_function.lambda_handler",
                Code={"ZipFile": zipped_code},
                Description=f"{self.app_id} tool: {self.name}",
                Timeout=30,
                MemorySize=256,
            )
        except ClientError as e:
            error = f"Error creating lambda. Client error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        except Exception as e:
            error = f"Error creating lambda. Unexpected error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        # Clean up the .zip archive build
        try:
            shutil.rmtree(self.temp_build_root)
        except OSError as e:
            error = (
                f"Error deleting temporary build directory {self.temp_build_root}: {e}"
            )
            self.logger.error(error)
            pass

        # Add the resourced-base policy allowing the lambda to be invoked
        # by the agent.
        try:
            add_permission_response = lambda_client.add_permission(
                StatementId=f"{self.app_id}-bedrock-invoke",
                FunctionName=response["FunctionArn"],
                Action="lambda:InvokeFunction",
                Principal="bedrock.amazonaws.com",
                # TODO: fix this. Remove wildcard
                SourceArn=f"arn:aws:bedrock:us-east-1:{self.aws_account_id}:agent/*",
            )
        except ClientError as e:
            error = f"Error attaching resource policy to lambda execution role. Client error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)
        except Exception as e:
            error = f"Error attaching resource policy to lambda execution role. Unexpected error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        return response["FunctionArn"]

    def _create_bedrock_lambda_role(self) -> str:
        """Get or create IAM role for Bedrock integration"""
        iam = boto3.client("iam")

        self.lambda_role_name = f"{self.app_id}_bedrock_lambda_execution_role"
        self.assume_role = self._get_assumed_role(service="lambda.amazonaws.com")

        try:
            iam.create_role(
                RoleName=self.lambda_role_name,
                AssumeRolePolicyDocument=json.dumps(self.assume_role),
            )
        except ClientError as e:
            error = f"Error creating lambda execution role. Client error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)
        except Exception as e:
            error = f"Error creating lambda execution role. Unexpected error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        try:
            iam.attach_role_policy(
                RoleName=self.lambda_role_name,
                PolicyArn="arn:aws:iam::aws:policy/service-role/AWSLambdaBasicExecutionRole",
            )
        except ClientError as e:
            error = f"Error attaching managed policy to lambda execution role. Client error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        except Exception as e:
            error = f"Error attaching managed policy to lambda execution role. Unexpected error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        return iam.get_role(RoleName=self.lambda_role_name)["Role"]["Arn"]

    def _create_bedrock_agent(self) -> str:
        """Create Bedrock agent"""

        self.agent_role_arn = self._create_agent_role()

        args = {
            "agentName": f"{self.app_id}_{self.agent_name}Agent",
            "agentResourceRoleArn": self.agent_role_arn,
            "foundationModel": self.agent_target_model,
            "idleSessionTTLInSeconds": self.agent_session_timeout,
            "instruction": self.agent_instruction,
            "description": self.agent_description,
        }

        try:
            response = self.bedrock_agent_client.create_agent(**args)
        except ClientError as e:
            error = f"Error creating agent. Client error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        except Exception as e:
            error = f"Error creating agent. Unexpected error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        return response["agent"]["agentId"]

    def _create_agent_action_group(self):
        """Create Bedrock action group and attach to Bedrock agent"""

        args = {
            "agentId": self.agent_id,
            "actionGroupExecutor": {
                "lambda": self.lambda_arn,
            },
            "actionGroupName": f"{self.app_id}_{self.agent_name}_ag",
            "agentVersion": "DRAFT",
            "apiSchema": {"payload": json.dumps(self.openapi_schema_dict)},
            "description": self.agent_description,
        }

        # Wait until agent status transitions to CREATING before creating the action group.
        counter = 0
        agent_details = self.bedrock_agent_client.get_agent(agentId=self.agent_id)
        agent_status = agent_details["agent"]["agentStatus"]
        while (agent_status == "CREATING") and (counter <= 9):
            counter = counter + 1
            time.sleep(1)
            agent_details = self.bedrock_agent_client.get_agent(agentId=self.agent_id)
            agent_status = agent_details["agent"]["agentStatus"]

        if counter == 10:
            error = f"Error creating agent timeout. Agent created, but status never transitioned out of CREATING"
            self.logger.error(error)
            raise RuntimeError(error)

        try:
            agent_action_group = self.bedrock_agent_client.create_agent_action_group(
                **args
            )
            self.agent_action_group_id = agent_action_group["agentActionGroup"][
                "actionGroupId"
            ]
        except ClientError as e:
            error = f"Error creating agent action group. Client error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)
        except Exception as e:
            error = f"Error creating agent action group. Unexpected error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        return agent_action_group

    def _create_agent_role(self) -> str:
        """Create IAM agent role for Bedrock integration"""

        iam = boto3.client("iam")

        self.agent_role_name = f"{self.app_id}_bedrock_agent_execution_role"
        self.agent_policy_name = f"{self.app_id}_bedrock_agent_allow_model"

        assume_role = self._get_assumed_role(service="bedrock.amazonaws.com")

        try:
            iam.create_role(
                RoleName=self.agent_role_name,
                AssumeRolePolicyDocument=json.dumps(assume_role),
            )
        except ClientError as e:
            error = f"Error creating agent role. Client error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)
        except Exception as e:
            error = f"Error creating agent role. Unexpected error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        managed_policy_txt = self._get_managed_policy(
            resource=f"foundation-model/{self.agent_target_model}"
        )

        try:
            managed_policy = iam.create_policy(
                PolicyName=self.agent_policy_name,
                PolicyDocument=json.dumps(managed_policy_txt),
            )
        except ClientError as e:
            error = f"Error creating agent policy. Client error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)
        except Exception as e:
            error = f"Error creating agent policy. Unexpected error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)

        self.agent_role_policy_arn = (
            f"arn:aws:iam::{self.aws_account_id}:policy/{self.agent_policy_name}"
        )
        try:
            iam.attach_role_policy(
                RoleName=self.agent_role_name,
                PolicyArn=self.agent_role_policy_arn,
            )
        except ClientError as e:
            error = f"Error creating attaching policy to agent role. Client error: {e}"
            self.logger.error(error)
            raise RuntimeError(error)
        except Exception as e:
            error = (
                f"Error creating attaching policy to agent role. Unexpected error: {e}"
            )
            self.logger.error(error)
            raise RuntimeError(error)

        return iam.get_role(RoleName=self.agent_role_name)["Role"]["Arn"]

    def _get_assumed_role(self, service: str = None) -> str:
        """Skeleton code for creating a role"""
        role = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": {"Service": f"{service}"},
                    "Action": "sts:AssumeRole",
                }
            ],
        }
        return role

    def _get_managed_policy(self, resource: str = None) -> str:
        """Skeleton code for creating a policy"""
        managed_policy_txt = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Action": "bedrock:InvokeModel",
                    "Resource": f"arn:aws:bedrock:us-east-1::{resource}",
                    "Effect": "Allow",
                }
            ],
        }

        return managed_policy_txt

    def _get_lambda_source(
        self, module_name: str = None, class_name: str = None
    ) -> str:
        """Bedrock wrapped lambda used by the agent"""

        # funky spacing below is required...
        lambda_src = f"""
import json
from {module_name} import {class_name}


def lambda_handler(event, context):
    tool = {class_name}()
    result = tool.fn()
    action_response = {{
        "actionGroup": event["actionGroup"],
        "apiPath": event["apiPath"],
        "httpMethod": event["httpMethod"],
        "httpStatusCode": 200,
        "responseBody": {{"application/json": {{"body": json.dumps(result)}}}},
    }}
    api_response = {{"messageVersion": "1.0", "response": action_response}}

    return api_response"""
        return lambda_src

    def delete_app(self, bedrock_config: dict = None) -> bool:
        """Tear down the boto3 stack created by .export_to_bedrock()"""

        valid_bedrock_config_schema = {
            "type": "object",
            "properties": {
                "lambdaARN": {"type": "string"},
                "lambdaRoleName": {"type": "string"},
                "agentId": {"type": "string"},
                "agentRoleName": {"type": "string"},
                "agentPolicyARN": {"type": "string"},
                "agentActionGroup": {"type": "string"},
            },
            "required": [
                "lambdaARN",
                "lambdaRoleName",
                "agentId",
                "agentRoleName",
                "agentPolicyARN",
                "agentActionGroup",
            ],
        }
        try:
            validate(bedrock_config, valid_bedrock_config_schema)
        except jsonschema.exceptions.ValidationError as err:
            error = f"JSON validation failure on input bedrock_config: {err}"
            self.logger.error(error)
            raise RuntimeError(error)

        self.lambda_client = boto3.client("lambda")
        self.iam_client = boto3.client("iam")
        self.bedrock_agent_client = boto3.client("bedrock-agent")

        # Order is important here...
        try:
            detach_agent_policy = self.iam_client.detach_role_policy(
                RoleName=bedrock_config["agentRoleName"],
                PolicyArn=bedrock_config["agentPolicyARN"],
            )
        except ClientError as e:
            error = f"Error detaching agent policy. Client error: {e}"
            self.logger.error(error)
            pass
        except Exception as e:
            error = f"Error detaching agent policy. Unexpected error: {e}"
            self.logger.error(error)
            pass
        try:
            delete_agent_role = self.iam_client.delete_role(
                RoleName=bedrock_config["agentRoleName"]
            )
        except ClientError as e:
            error = f"Error deleting agent role. Client error: {e}"
            self.logger.error(error)
            pass
        except Exception as e:
            error = f"Error deleting agent role. Unexpected error: {e}"
            self.logger.error(error)
            pass

        try:
            delete_agent_policy = self.iam_client.delete_policy(
                PolicyArn=bedrock_config["agentPolicyARN"]
            )
        except ClientError as e:
            error = f"Error deleting agent policy. Client error: {e}"
            self.logger.error(error)
            pass
        except Exception as e:
            error = f"Error deleting agent policy. Unexpected error: {e}"
            self.logger.error(error)
            pass

        try:
            delete_agent = self.bedrock_agent_client.delete_agent(
                agentId=bedrock_config["agentId"], skipResourceInUseCheck=True
            )
        except ClientError as e:
            error = f"Error deleting agent. Client error: {e}"
            self.logger.error(error)
            pass
        except Exception as e:
            error = f"Error deleting agent. Unexpected error: {e}"
            self.logger.error(error)
            pass

        try:
            detach_lambda_policy = self.iam_client.detach_role_policy(
                RoleName=bedrock_config["lambdaRoleName"],
                PolicyArn="arn:aws:iam::aws:policy/service-role/AWSLambdaBasicExecutionRole",
            )
        except ClientError as e:
            error = f"Error detaching lambda policy. Client error: {e}"
            self.logger.error(error)
            pass
        except Exception as e:
            error = f"Error detaching lambda policy. Unexpected error: {e}"
            self.logger.error(error)
            pass

        try:
            delete_lambda_role = self.iam_client.delete_role(
                RoleName=bedrock_config["lambdaRoleName"]
            )
        except ClientError as e:
            error = f"Error deleting lambda role. Client error: {e}"
            self.logger.error(error)
            pass
        except Exception as e:
            error = f"Error deleting lambda role. Unexpected error: {e}"
            self.logger.error(error)
            pass

        try:
            delete_lambda = self.lambda_client.delete_function(
                FunctionName=bedrock_config["lambdaARN"]
            )
        except ClientError as e:
            error = f"Error deleting lambda. Client error: {e}"
            self.logger.error(error)
            pass
        except Exception as e:
            error = f"Error deleting lambda. Unexpected error: {e}"
            self.logger.error(error)
            pass



from .base import HeadlessBrowserBase
from ..config import FunctionRegistry
import logging

logger = logging.getLogger(__name__)

@FunctionRegistry.register
class HeadlessBrowserGet(HeadlessBrowserBase):
    def __init__(self, provider="selenium-chrome", name="headless_browser_get"):
        super().__init__(provider=provider)
        self.name = name

    @property
    def definition(self):
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": "Retrieve the contents of a web page with JavaScript rendered using a headless browser.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "url": {
                            "type": "string",
                            "description": "The URL of the web page to fetch"
                        }
                    },
                    "required": ["url"]
                }
            }
        }

    def fn(self, url):
        logger.debug(f"Fetching URL with headless browser using provider {self.provider}: {url}")
        return self.get_page_source(url)


try:
    from langflow.custom import Component
    from langflow.io import MessageTextInput, IntInput, BoolInput, FloatInput, Output
    from langflow.schema import Data

    _HAS_LANGFLOW = True
except ImportError:
    _HAS_LANGFLOW = False


class LangflowMixin:
    def import_from_langflow(self, langflow_component):
        """Adapt a Langflow component to a Gofannon tool"""
        if not _HAS_LANGFLOW:
            raise RuntimeError(
                "langflow is not installed. Install with `pip install langflow`"
            )

        self.name = langflow_component.display_name.replace(" ", "_").lower()
        self.description = langflow_component.description

        # Extract parameters from component inputs
        parameters = {
            "type": "object",
            "properties": {},
            "required": []
        }

        for component_input in langflow_component.inputs:
            if component_input.name in ["self", "context"]:
                continue

            param_type = "string"
            if isinstance(component_input, (IntInput, FloatInput)):
                param_type = "number"
            elif isinstance(component_input, BoolInput):
                param_type = "boolean"

            parameters["properties"][component_input.name] = {
                "type": param_type,
                "description": component_input.info or ""
            }

            if component_input.required:
                parameters["required"].append(component_input.name)

        self.definition = {
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": parameters
            }
        }

        # Create execution wrapper
        def adapted_fn(**kwargs):
            result = langflow_component.build()(**kwargs)
            return result.data if isinstance(result, Data) else result

        self.fn = adapted_fn

    def export_to_langflow(self):
        """Convert Gofannon tool to Langflow component"""
        if not _HAS_LANGFLOW:
            raise RuntimeError(
                "langflow is not installed. Install with `pip install langflow`"
            )

        parameters = self.definition.get("function", {}).get("parameters", {})
        param_properties = parameters.get("properties", {})
        required_params = parameters.get("required", [])

        # Create input fields
        component_inputs = []
        type_map = {
            "string": MessageTextInput,
            "number": FloatInput,
            "integer": IntInput,
            "boolean": BoolInput
        }

        for param_name, param_def in param_properties.items():
            input_type = param_def.get("type", "string")
            InputClass = type_map.get(input_type, MessageTextInput)

            component_inputs.append(
                InputClass(
                    name=param_name,
                    display_name=param_name.replace("_", " ").title(),
                    info=param_def.get("description", ""),
                    required=param_name in required_params,
                    tool_mode=True
                )
            )

            # Define component class
        class ExportedComponent(Component):
            display_name = self.definition["function"]["name"].title()
            description = self.definition["function"]["description"]
            icon = "tool"

            inputs = component_inputs
            outputs = [Output(display_name="Result", name="result", method="run_tool")]

            def run_tool(self, **kwargs):
                result = self.tool.execute(context=None, **kwargs)
                return Data(data=result.output)

                # Attach tool instance to component class
        ExportedComponent.tool = self
        return ExportedComponent


import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Callable
import json
import logging
from pathlib import Path

import anyio

from ..config import ToolConfig

from .smol_agents import SmolAgentsMixin
from .langchain import LangchainMixin
from .bedrock import BedrockMixin
from .langflow import LangflowMixin
from .mcp import MCPMixin
from .llamastack import LlamaStackMixin


@dataclass
class ToolResult:
    success: bool
    output: Any
    error: str = None
    retryable: bool = False


class WorkflowContext:
    def __init__(self, firebase_config=None):
        self.data = {}
        self.execution_log = []
        self.firebase_config = firebase_config
        self.local_storage = Path.home() / ".llama" / "checkpoints"
        self.local_storage.mkdir(parents=True, exist_ok=True)

    def save_checkpoint(self, name="checkpoint"):
        if self.firebase_config:
            self._save_to_firebase(name)
        else:
            self._save_local(name)

    def _save_local(self, name):
        path = self.local_storage / f"{name}.json"
        with open(path, "w") as f:
            json.dump({"data": self.data, "execution_log": self.execution_log}, f)

    def _save_to_firebase(self, name):
        from firebase_admin import firestore

        db = firestore.client()
        doc_ref = db.collection("checkpoints").document(name)
        doc_ref.set(
            {
                "data": self.data,
                "execution_log": self.execution_log,
                "timestamp": firestore.SERVER_TIMESTAMP,
            }
        )

    def log_execution(self, tool_name, duration, input_data, output_data):
        entry = {
            "tool": tool_name,
            "duration": duration,
            "input": input_data,
            "output": output_data,
        }
        self.execution_log.append(entry)


class BaseTool(SmolAgentsMixin,
               LangchainMixin,
               BedrockMixin,
               LangflowMixin,
               MCPMixin,
               LlamaStackMixin,
               ABC):
    def __init__(self, **kwargs):
        self.logger = logging.getLogger(
            f"{self.__class__.__module__}.{self.__class__.__name__}"
        )
        self._load_config()
        self._configure(**kwargs)
        self.logger.debug("Initialized %s tool", self.__class__.__name__)

    def _configure(self, **kwargs):
        """Set instance-specific configurations"""
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def _load_config(self):
        """Auto-load config based on tool type"""
        if hasattr(self, "API_SERVICE"):
            self.api_key = ToolConfig.get(f"{self.API_SERVICE}_api_key")

    @property
    @abstractmethod
    def definition(self):
        pass

    @property
    def output_schema(self):
        return self.definition.get("function", {}).get("parameters", {})

    @abstractmethod
    def fn(self, *args, **kwargs):
        pass

    def execute(self, context: WorkflowContext, **kwargs) -> ToolResult:
        try:
            start_time = time.time()
            result = self.fn(**kwargs)
            duration = time.time() - start_time

            context.log_execution(
                tool_name=self.__class__.__name__,
                duration=duration,
                input_data=kwargs,
                output_data=result,
            )

            return ToolResult(success=True, output=result)
        except Exception as e:
            return ToolResult(success=False, output=None, error=str(e), retryable=True)

    async def execute_async(self, arguments: dict):
        return await anyio.to_thread.run_sync(self.fn, **arguments)



from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from ..base import BaseTool
import logging

logger = logging.getLogger(__name__)

class HeadlessBrowserBase(BaseTool):
    """
    Base class for headless browser tools.
    On initialization, the provider parameter may be one of:
    "selenium-chrome", "selenium-firefox", "lightpanda", or "remote".
    Currently, only "selenium-chrome" is supported.
    """
    def __init__(self, provider="selenium-chrome", **kwargs):
        super().__init__(**kwargs)
        self.provider = provider.lower()
        supported = ["selenium-chrome", "selenium-firefox", "lightpanda", "remote"]
        if self.provider not in supported:
            raise ValueError(f"Unsupported provider: {self.provider}. Supported providers: {supported}")

    def _get_driver(self):
        if self.provider == "selenium-chrome":
            chrome_options = Options()
            chrome_options.add_argument("--headless")
            chrome_options.add_argument("--disable-gpu")
            chrome_options.add_argument("--no-sandbox")
            # Adjust the executable_path if needed or ensure the chromedriver is in PATH.
            driver = webdriver.Chrome(options=chrome_options)
            return driver
        else:
            raise NotImplementedError(f"Provider {self.provider} is not supported in the current implementation.")

    def get_page_source(self, url):
        driver = self._get_driver()
        try:
            driver.get(url)
            # Optionally, add waits here if needed so that JavaScript can fully execute.
            page_source = driver.page_source
        finally:
            driver.quit()
        return page_source